/* 
 * $Id: bgp_dump.c,v 1.24 1999/02/08 21:31:23 labovit Exp $
 */

#include <mrt.h>
#include <bgp.h>
#include <sys/wait.h>

/* local functions */
static void dump_binary_bgp_view (int fd, view_t *view); 
static void dump_routing_table (time_t now, int viewnow, char *sync, int *synclen,
				enum MRT_MSG_BGP_TYPES bgptype);

void
bgp_write_mrt_msg (bgp_peer_t *peer, enum MRT_MSG_BGP_TYPES bgptype, 
		   u_char *buf, int length) {
    u_char tmpx[MAX_MSG_SIZE], *cp = tmpx;
    int len = length;
    char sync[MAXLINE];
    int synclen = 0;
    time_t now;
    enum MRT_MSG_TYPES msgtype = MSG_PROTOCOL_BGP;
    int viewno = peer->view;
#if defined(_REENTRANT) && defined(HAVE_LOCALTIME_R)
    struct tm my_tm;
#endif

    if (BGP->dump_update_form == NULL && BGP->dump_route_form[viewno] == NULL)
	return;

    pthread_mutex_lock (&BGP->mutex_lock);
    time (&now);

    /* Dump binary/ascii routing table */
    if (BGP->dump_route_form[viewno]) {
      if (BGP->dump_route_time[viewno] == 0) {
	BGP->dump_route_time[viewno] = now;
      }
      else if ((BGP->dump_route_time[viewno] + BGP->dump_route_interval[viewno]) < now) {
	trace (NORM, MRT->trace, "About to start dump (%d past dump time)\n",
	       now - (BGP->dump_route_time[viewno] + BGP->dump_route_interval[viewno]));
	dump_routing_table(now, viewno, sync, &synclen, bgptype);
      }
    }

    /* dump update packet to disk */
    if (BGP->dump_update_form) {
        static io_t *IO = NULL;

        if (IO == NULL) {
	    IO = New_IO (NULL);
        }
  
        if (BGP->dump_update_time == 0 ||
	   (BGP->dump_update_interval && 
		BGP->dump_update_time + BGP->dump_update_interval <= now)) {

	    char name[MAXLINE];
	    struct tm *tm;
	    static int nerror = 0;

#if defined(_REENTRANT) && defined(HAVE_LOCALTIME_R)
	    tm = localtime_r (&now, &my_tm);
#else
	    tm = localtime (&now);
#endif
            strftime (name, sizeof (name), BGP->dump_update_form, tm);
	    if (io_set (IO, IO_OUTFILE, name, NULL) < 0) {
		if (nerror++ <= 0) {
		    trace (TR_ERROR, MRT->trace, 
		           "can not open dump file: %s (%s)\n",
		           name, strerror (errno));
		}
		else if (nerror >= 100) {
		    trace (TR_ERROR, MRT->trace, 
		           "give up opening dump file: %s (%s)\n",
		           name, strerror (errno));
		    Delete (BGP->dump_update_form);
    		    BGP->dump_update_form = NULL;
		}
		BGP->dump_update_time = now;
    		pthread_mutex_unlock (&BGP->mutex_lock);
		return;
	    }
	    else {
		nerror = 0;
		BGP->dump_update_time = now;
	    }
	}

        /* from */
        BGP_PUT_SHORT (peer->gateway->AS, cp);

#ifdef HAVE_IPV6
        if (peer->gateway->prefix->family == AF_INET6) {
            memcpy (cp, prefix_tochar (peer->gateway->prefix), 16);
            cp += 16;
        }
        else
#endif /* HAVE_IPV6 */
        BGP_PUT_NETLONG (prefix_tolong (peer->gateway->prefix), cp);

	/*
	 * To be compatible with Craig's code, AS and addr for 'to' 
	 * should not be included in this MRT message at the moment. 
	 *   -- masaki
	 */
	if (bgptype != MSG_BGP_STATE_CHANGE) {

	  /* to */
	  BGP_PUT_SHORT (BGP->my_as, cp);
#ifdef HAVE_IPV6
	  if (peer->gateway->prefix->family == AF_INET6) {
            if (peer->local_addr) {
	      assert (peer->local_addr->family == AF_INET6);
	      memcpy (cp, prefix_tochar (peer->local_addr), 16);
	    }
	    else {
	      memset (cp, 0, 16);
	    }
            cp += 16;
	  }
	  else
#endif /* HAVE_IPV6 */
	    if (peer->local_addr)
	      BGP_PUT_NETLONG (prefix_tolong (peer->local_addr), cp);
	    else
	      BGP_PUT_NETLONG (0L, cp);
	}

        memcpy (cp, buf, length);
        len += (cp - tmpx);

#ifdef HAVE_IPV6
        if (peer->gateway->prefix->family == AF_INET6) {
	  if (BIT_TEST (peer->options, BGP_BGP4PLUS_01))
	    msgtype = MSG_PROTOCOL_BGP4PLUS_01;
	  else
	    msgtype = MSG_PROTOCOL_BGP4PLUS;
        }
#endif /* HAVE_IPV6 */


#ifdef HAVE_PTHREAD_H
 	/* flush sync info */
	if (synclen > 0)
	  io_write (IO, 0, msgtype, MSG_BGP_SYNC, synclen, (char *)sync);
#endif /* HAVE_PTHREAD_H */

        io_write (IO, 0, msgtype, bgptype, len, (char *)tmpx);
    }
    pthread_mutex_unlock (&BGP->mutex_lock);
}


void
bgp_write_status_change (bgp_peer_t *peer, u_short state) {
    u_char buf[2+2], *cp = buf;

    BGP_PUT_SHORT (peer->state, cp); /* previous */
    BGP_PUT_SHORT (state, cp); /* new */
    bgp_write_mrt_msg (peer, MSG_BGP_STATE_CHANGE, buf, 4);
}



/* dump_binary_bgp_view 
 * 4 time stamp | 2 type | 2 subtype (familly) | 4 length
 * 2  view # | 
 *
 *  4 prefix | 1 mask | status (VRTS_SUPPRESS) | 4 time originated |
 *  4 len | attributes 
 * 
 */
static void dump_binary_bgp_view (int fd, view_t *view) {
  bgp_route_head_t *rt_head;
  bgp_route_t *route;
  u_char buffer[MAX_MSG_SIZE];
  time_t now;
  u_char *cp, *end, *start_attr;
  u_char *p_save_len, *p_total_attrib_len;
  int len;
  short seq_num = 0;
  
  time (&now);

  cp = buffer;
  end = buffer + MAX_MSG_SIZE;

  BGP_PUT_LONG (now, cp);		/* time */
  BGP_PUT_SHORT (MSG_TABLE_DUMP, cp);
  BGP_PUT_SHORT (0, cp);		/* subtype. maybe family ? */
  p_save_len = cp;
  BGP_PUT_LONG (0, cp);			/* skip length for now */

  BGP_PUT_SHORT (23, cp);		/* view */
  BGP_PUT_SHORT (seq_num, cp);		/* sequence number */
  seq_num++;

  VIEW_RADIX_WALK (view, rt_head) {

    if ((end - cp) < 500) {
      len = cp - buffer - 12;
      BGP_PUT_LONG (len, p_save_len);
      cp = buffer;
      write (fd, cp, len+12);
      trace (NORM, MRT->trace, 
	     "Dumping routing table (%d bytes) to disk\n", len+12);
      
      cp = buffer;
      memset (cp, 0, MAX_MSG_SIZE);
      
      BGP_PUT_LONG (now, cp);
      BGP_PUT_SHORT (MSG_TABLE_DUMP, cp);
      BGP_PUT_SHORT (0, cp);	/* subtype. maybe family ? */
      p_save_len = cp;
      BGP_PUT_LONG (0, cp);	/* skip length for now */

      BGP_PUT_SHORT (0, cp);	/* view */
      BGP_PUT_SHORT (seq_num, cp); /* view */
      seq_num++;
    }
      
    
    /* 
     * BGP Routes 
     */
    LL_Iterate (rt_head->ll_routes, route) {
      assert (rt_head->active);

      BGP_PUT_ADDR (prefix_tochar (rt_head->prefix), cp);
      BGP_PUT_BYTE (rt_head->prefix->bitlen, cp);

      /* XXX I need to check this Craig's code.
	 Probably it's OK putting state_bits but it's u_long. masaki */
      /* status */
      if (BIT_TEST (rt_head->state_bits, VRTS_SUPPRESS))
	BGP_PUT_BYTE (VRTS_SUPPRESS, cp);
      else if (BIT_TEST (rt_head->state_bits, VRTS_DELETE))
	BGP_PUT_BYTE (VRTS_DELETE, cp);
      else
	BGP_PUT_BYTE (1, cp); /* active ? */

      /* age, or really time last changed */
      BGP_PUT_LONG (route->time, cp);

      /* peer */
      BGP_PUT_ADDR (prefix_tochar (route->attr->gateway->prefix), cp);
      BGP_PUT_SHORT (route->attr->gateway->AS, cp);

      /* skip attribute length for now */
      p_total_attrib_len = cp;
      BGP_PUT_SHORT (0, cp);
      start_attr = cp;

      cp = bgp_add_attributes (cp, end - cp, route->attr, NULL);


      BGP_PUT_SHORT (cp - start_attr, p_total_attrib_len);
    }
  }
  VIEW_RADIX_WALK_END;

  /* does this include header ? */
  len = cp - buffer - 12;
  BGP_PUT_LONG (len, p_save_len);
  cp = buffer;
  write (fd, cp, len + 12);
  trace (NORM, MRT->trace, 
	 "Dumping routing table (%d bytes) to disk\n", len+12);
}


/*
 * dump_routing_table
 */
static void dump_routing_table (time_t now, int viewno, char *sync, 
				int *synclen, enum MRT_MSG_BGP_TYPES bgptype) {
  char name[MAXLINE];
  struct tm *tm;
  int fd;
  uii_connection_t uii;
  static int nerror = 0;
#ifndef HAVE_PTHREAD_H
  pid_t pid;
#endif /* HAVE_PTHREAD_H */

  BGP->dump_route_time[viewno] = now;

  tm = localtime (&now);
  strftime (name, sizeof (name), BGP->dump_route_form[viewno], tm);
  if ((fd = open (name, (O_CREAT | O_APPEND | O_WRONLY), 0666)) < 0) {
    if (nerror++ <= 0) {
      trace (TR_ERROR, MRT->trace, "can not open dump file: %s (%s)\n",
	     name, strerror (errno));
    }
    else if (nerror >= 100) {
      trace (TR_ERROR, MRT->trace, "give up opening dump file: %s (%s)\n",
	     name, strerror (errno));
      Delete (BGP->dump_route_form[viewno]);
      BGP->dump_route_form[viewno] = NULL;
    }
    return;
  }

  /* open of dump file succeeded ! */

  /* set sync information */
  if (BGP->dump_update_form) {
    char *xp = sync;
    BGP_PUT_SHORT (viewno, xp);
    strcpy (xp, name);
    *synclen = sizeof (short) + strlen (name) + 1 /* for zero */;
  }

#ifndef HAVE_PTHREAD_H
  /* it was taking too long to do the dumps, hopefully forking
   * will be faster...
   */

  if ((pid = fork ()) < 0) {
    trace (TR_ERROR, MRT->trace, "Fork error in bgp_dump.c (%d)\n", pid);
    return;
  }
  else if (pid > 0) {	/* parent -- main process */
    trace (NORM, MRT->trace, "Parent before waitpid (%d)\n", pid);

    if (waitpid (pid, NULL, 0) != pid)
      perror ("waitpid error");

    trace (NORM, MRT->trace, "Parent after waitpid. Now returning...\n");
    return;
  }

  set_thread_id (pid);
  trace (NORM, MRT->trace, "Forked once %d\n", pid);

  /* so pid == 0, first child 
   * fork again so that not being a process group leader  
   */
  if ((pid = fork ()) < 0) {
    trace (TR_ERROR, MRT->trace, "Error in second fork in bgp_dump.c (%d)\n", pid);
    exit (1);
  }
  else if (pid > 0) {	/* parent of second fork */
    exit (0);
  }

  /* okay, so now only orphaned child should be proceeding on */
  set_thread_id (pid);
  trace (NORM, MRT->trace, "Forked twice %d\n. I am dump grandchild...\n", pid);


#endif /* not HAVE_PTHREAD_H */

  lseek (fd, 0, SEEK_END);
  memset (&uii, 0, sizeof (uii));
  uii.sockfd = -1;
  uii.sockfd_out = fd;
  /* automated */
  if (bgptype == MSG_BGP_UPDATE) {
    if (BGP->dump_route_type == DUMP_ASCII) {
      bgp_dump_view (&uii, BGP->views[viewno], NULL, NULL, NULL);
      uii_send_bulk_data (&uii);
    }
    else
      dump_binary_bgp_view (fd, BGP->views[viewno]);
  }
  else {
    /* requested by UII */
    /* MASAKI -- why do we view_open here, but not above??? */
    view_open (BGP->views[viewno]);
    if (BGP->dump_route_type == DUMP_ASCII) {
      bgp_dump_view (&uii, BGP->views[viewno], 
		     NULL, NULL, NULL);
      uii_send_bulk_data (&uii);
    }
    else
      dump_binary_bgp_view (fd, BGP->views[viewno]);
    view_close (BGP->views[viewno]);
  }
  close (fd);

  
#ifndef HAVE_PTHREAD_H
  /* we are the child, exit here */
  trace (NORM, MRT->trace, "Dump child exiting..\n");

  exit (0);
#endif /* HAVE_PTHREAD_H */
}

